#include "syscall.h"
#include "config.h"
#include "occluder.h"
#include <set>
#include <GL/glut.h>
#include <mpi.h>
#include <stdio.h>
#include <math.h>

using namespace std;

#define min(x, y)   (x>y?y:x)
#define max(x, y)   (x>y?x:y)

#define RangeMin(x, low, high) (max(min((x), (high)), (low)))
#define RangeMax(x, low, high) (min(max((x), (low)), (high)))

void Occluder::transformVertex(MatrixType transformer, FVector vertex, FVector result)
{
	float src[4], dest[4];
	int i, j;
	for (i=0; i<3; i++)
	{
		src[i]=vertex[i];
	}
	src[3]=1;

	for (i=0; i<4; i++)
	{
		dest[i]=0;
		for (j=0; j<4; j++)
		{
			dest[i]+=src[j]*transformer[i+j*4];
		}		
	}
	for (i=0; i<3; i++)
	{
		result[i]=dest[i]/dest[3];
	}
}

void Occluder::matrixMultiply(MatrixType left, MatrixType right, MatrixType result)
{
	int i, j, k;

	for (i=0; i<4; i++)
	{
		for (j=0; j<4; j++)
		{
			result[i*4+j]=0;
			for (k=0; k<4; k++)
			{
				result[i*4+j]+=left[i*4+k]*right[k*4+j];
			}
			
		}
	}
}

void printMatrix(MatrixType matrix)
{
	for (int i=0; i<4; i++)
	{
		for (int j=0; j<4; j++)
		{
			printf("%f,", matrix[i*4+j]);
		}
		printf("\n");
	}
	printf("\n");
}

void Occluder::calcTransform()
{
	int i;
	MatrixType modelMatrix, projectMatrix, tempMatrix, viewMatrix;
	int viewVector[4];

	glGetFloatv(GL_MODELVIEW_MATRIX, modelMatrix);
	glGetFloatv(GL_PROJECTION_MATRIX, projectMatrix);
	glGetIntegerv(GL_VIEWPORT, viewVector);


	matrixMultiply(modelMatrix, projectMatrix, tempMatrix);

	for (i=0; i<16; i++)
	{
		viewMatrix[i]=0;
	}
	viewMatrix[0]=viewVector[2]/2;
	viewMatrix[12]=viewVector[2]/2;
	viewMatrix[5]=-viewVector[3]/2;
	viewMatrix[13]=viewVector[3]/2;
	viewMatrix[10]=1;
	viewMatrix[15]=1;
	matrixMultiply(tempMatrix, viewMatrix, transformer);
}

Occluder::Occluder()
{
	occluderComm=MPI_COMM_NULL;
}	

void Occluder::init(int rank)
{
	MPI_Group commGroup;
	int testerRanks[TesterCount];
	int i;
	char logName[32];
	myTesterRank=rank;
	for (i=0; i<TesterCount; i++)
        {
                testerRanks[i]=TesterHostStartIndex+i;
        }
	
        MPI_Comm_group(MPI_COMM_WORLD, &commGroup);
        MPI_Group_incl(commGroup, TesterCount, testerRanks, &occluderGroup);
        MPI_Comm_create(MPI_COMM_WORLD, occluderGroup, &occluderComm);
	myRegion.rect.topLeft.x=myRegion.rect.topLeft.y=0;
	myRegion.rect.bottomRight.x=myRegion.rect.bottomRight.y=ScreenSize-1;
	myRegion.maxz=1.0;
	sprintf(logName, "occluderLog%d.txt", rank);
	log=fopen(logName, "w");
	CHECK_NEW(log);
}

void Occluder::uninit()
{
	fclose(log);
}

FVector operations[8]=
{
	{1,1,1},
	{1,1,-1},
	{-1,1,-1},
	{-1,1,1},
	{1,-1,1},
	{1,-1,-1},
	{-1,-1,-1},
	{-1,-1,1},
};


/*
bool Occluder::testRegion(Rect& rect, float& minDepth)
{
	Region rgn;
	RegionSet::iterator it;
	pair<RegionSet::iterator, RegionSet::iterator> result;
	rgn.rect=rect;
	result=regionSet.equal_range(rgn);
	
	it=result.first;
	if (it==regionSet.end())
	{
		return false;
	}

	while (it!=result.second)
	{
		if (rect.isInside(it->rect))
		{
			if (minDepth>=it->maxz)
			{
				return true;
			}
		}
		it++;
	}
	return false;
}
*/


bool Occluder::testRegion(Rect& rect, float& minDepth)
{
	Region rgn;
	RegionSet::iterator it;
	rgn.rect=rect;
	it=regionSet.lower_bound(rgn);
	
	while (it!=regionSet.end())
	{
		if (rect.isInside(it->rect))
		{
			if (minDepth>=it->maxz)
			{
				return true;
			}
		}
		it++;
	}
	return false;
}


void Occluder::outputRegionSet()
{
	RegionSet::iterator it;
	for (it=regionSet.begin(); it!=regionSet.end(); ++it)
	{
		fprintf(log, "rect[%d,%d][%d,%d]depth[%f]\n", it->rect.topLeft.x, 
			it->rect.topLeft.y, it->rect.bottomRight.x, it->rect.bottomRight.y,
			it->maxz);
	}
}

void Occluder::doCalcRegion(VVector src, Rect& rect, float& depth)
{
	VVector dest;
	int x, y;
	transformVertex(transformer, src, dest);
	x=floor(dest[0]);
	y=floor(dest[1]);
	rect.topLeft.x = RangeMin(x, 0, rect.topLeft.x);
	rect.topLeft.y = RangeMin(y, 0, rect.topLeft.y);
	x=ceil(dest[0]);
	y=ceil(dest[1]);

	rect.bottomRight.x = RangeMax(x, rect.bottomRight.x, ScreenSize - 1);
	rect.bottomRight.y = RangeMax(y, rect.bottomRight.y, ScreenSize - 1);
	depth=min(depth, dest[2]);
}

void Occluder::initCalcRegion(Rect& rect, float& depth)
{
	rect.topLeft.x=rect.topLeft.y=ScreenSize-1;
	rect.bottomRight.x=rect.bottomRight.y=0;
	depth=1.0;
}

bool Occluder::calcRegion(PlyVertex* vPtr, UIVector* fPtr, int faceNumber, Rect& rect, float& depth)
{
	int index;
	initCalcRegion(rect, depth);
	for (int i=0; i<faceNumber; i++)
	{
		for (int j=0; j<3; j++)
		{
			index = fPtr[i][j];
			doCalcRegion(vPtr[index].coord, rect, depth);
		}
	}
	return testCalcRegion(rect);
}				
	
int Occluder::doOccluderTest(PlyVertex* vPtr, UIVector* fPtr, int faceNumber)
{
	int index, result = 0;
	Rect rect;
	float depth;

	for (int i = 0; i < faceNumber; i ++)
	{
		initCalcRegion(rect, depth);
		for (int j = 0; j < 3; j ++)
		{
			index = fPtr[i][j];
			doCalcRegion(vPtr[index].coord, rect, depth);
		}
		if (checkRegion(rect, depth))
		{
			//fprintf(log, "region[%d,%d,%d,%d][%f]\n", rect.topLeft.x, rect.topLeft.y, rect.bottomRight.x, 
			//	rect.bottomRight.y, depth);
			result ++;
		}
	}
	return result;
}			
	
bool Occluder::testCalcRegion(Rect& rect)
{
	if (rect.bottomRight.x<rect.topLeft.x||rect.bottomRight.y<rect.topLeft.y||rect.topLeft.x<0||rect.topLeft.y<0)
	{
		//it is empty
		return false;
	}
	return true;
}

bool Occluder::calcRegion(FVector center, float width, Rect& rect, float& depth)
{
	int i, j;
	int x, y;
	FVector src, dest;
	initCalcRegion(rect, depth);
	for (i=0; i<8; i++)
	{
		for (j=0; j<3; j++)
		{
			src[j]=center[j]+operations[i][j]*width;
		}
		doCalcRegion(src, rect, depth);
	}
	return testCalcRegion(rect);
}		 


void Occluder::readRegion(Rect& rect)
{
	myRegion.rect=rect;
	glReadPixels(myRegion.rect.topLeft.x, myRegion.rect.topLeft.y, myRegion.rect.bottomRight.x-
		myRegion.rect.topLeft.x+1, myRegion.rect.bottomRight.y-myRegion.rect.topLeft.y+1,
		GL_DEPTH_COMPONENT, GL_FLOAT, depthBuffer);	
}



/*
void Occluder::analysisRegion()
{
	int width, height;
	float curMax=0;
	int w, h;
	int x, y, curW, curH;
	width=myRegion.rect.bottomRight.x-myRegion.rect.topLeft.x;
	height=myRegion.rect.bottomRight.y-myRegion.rect.topLeft.y;
	x=myRegion.rect.topLeft.x;
	y=myRegion.rect.topLeft.y;
	curW=curH=0;

	fprintf(log, "width=%d,height=%d\n", width, height);
	fflush(log);

	//let's stick to a simple one first
	for (h=0; h<height; h++)
	{
		for (w=0; w<width; w++)
		{
			if (depthBuffer[w+h*width]>curMax)
			{
				curMax=depthBuffer[w+h*width];
			}
			if (curMax==1.0)
			{
				break;
			}
		}
		if (curMax==1.0)
		{
			break;
		}
	}
	myRegion.maxz=curMax;
}					
*/


/*
void Occluder::findNextRect(const MyPoint& origin, int width, int height, RectVector& rectVector)
{
	MyPoint start;
	start=origin;

	while (start.x<width&&start.y<height)
	{
		if (depthBuffer[start.y*width+start.x]==1.0)
		{
			start.x++;
			if (start.x>=width)
			{
				start.y++;
				start.x=origin.x;
			}
		}
	
		
		
	
	

void Occluder::findEnd(const MyPoint& start, int width, int height, MyPoint& end)
{
	int w=0, h=0;

	//first find col
	end.x=width-1;
	end.y=height-1;
	for (w=start.x; w<width; w++)
	{
		if (depthBuffer[start.y*width+w]==1.0)
		{
			end.x=w;
			break;
		}
	}
	
	for (h=start.y; h<height; h++)
	{
		for (w=start.x; w<end.x; w++)
		{
			if (depthBuffer[h*width+w]==1.0)
			{
				end.y=h;
				return;
			}
		}
	}
	
}
				



void Occluder::findRegion(MyPoint start, MyPoint end,  int& width, int& height)
{
	int w=0, h=0;
	while (h+start.y<=end.y && w+start.x<=end.x)
	{
		
		while (depthBuf[
		
		
*/		
		
void Occluder::analysisRegion()
{
	int width, height;
	float curMax=0;
	int w, h;
	int x, y, curW, curH;
	width=myRegion.rect.bottomRight.x-myRegion.rect.topLeft.x;
	height=myRegion.rect.bottomRight.y-myRegion.rect.topLeft.y;
	x=myRegion.rect.topLeft.x;
	y=myRegion.rect.topLeft.y;
	curW=curH=0;

	fprintf(log, "width=%d,height=%d\n", width, height);
	fflush(log);

	//let's stick to a simple one first
	for (h=0; h<height; h++)
	{
		for (w=0; w<width; w++)
		{
			if (depthBuffer[w+h*width]>curMax)
			{
				curMax=depthBuffer[w+h*width];
			}
			if (curMax==1.0)
			{
				break;
			}
		}
		if (curMax==1.0)
		{
			break;
		}
	}
	myRegion.maxz=curMax;
}					



bool Occluder::intersectRegion(Region* ptr, RegionSet::iterator it)
{
	Region rgn[3];
	RegionSet::iterator posIt;
	pair<RegionSet::iterator, bool> result;	

	if (ptr->rect.topLeft.x<it->rect.topLeft.x)
	{
		rgn[0].rect.topLeft=ptr->rect.topLeft;
		rgn[0].rect.bottomRight.x=it->rect.topLeft.x;
		rgn[0].rect.bottomRight.y=ptr->rect.bottomRight.y;
		rgn[0].maxz=ptr->maxz;

		//done with rgn[0]
		rgn[1].rect.topLeft.x=it->rect.topLeft.x;

		rgn[2].maxz=it->maxz;
	}
	else
	{
		rgn[0].rect.topLeft=it->rect.topLeft;
		rgn[0].rect.bottomRight.x=ptr->rect.topLeft.x;
		rgn[0].rect.bottomRight.y=it->rect.bottomRight.y;
		rgn[0].maxz=it->maxz;
		//done with rgn[0]
		rgn[1].rect.topLeft.x=ptr->rect.topLeft.x;
		rgn[2].maxz=ptr->maxz;
	}

	if (ptr->rect.bottomRight.x>it->rect.bottomRight.x)
	{
		rgn[2].rect.bottomRight=ptr->rect.bottomRight;
		rgn[2].rect.topLeft.x=it->rect.bottomRight.x;
		rgn[2].rect.topLeft.y=ptr->rect.topLeft.y;
		//done with rgn[2]
		rgn[1].rect.bottomRight.x=it->rect.bottomRight.x;
	}
	else
	{
		rgn[2].rect.bottomRight=it->rect.bottomRight;
		rgn[2].rect.topLeft.x=ptr->rect.bottomRight.x;
		rgn[2].rect.topLeft.y=it->rect.topLeft.y;
		//done with rgn[2]
		rgn[1].rect.bottomRight.x=ptr->rect.bottomRight.x;
	}

	rgn[1].rect.topLeft.y=min(ptr->rect.topLeft.y, it->rect.topLeft.y);

	rgn[1].rect.bottomRight.y=max(ptr->rect.bottomRight.y, it->rect.bottomRight.y);
	
	rgn[1].maxz=max(rgn[0].maxz, rgn[2].maxz);
	regionSet.erase(it);
	for (int i=0; i<3; i++)
	{
		result=regionSet.insert(rgn[i]);
		if (!result.second)
		{
			fprintf(log, "begin recursively intersect region\n");
			fflush(log);
			intersectRegion(rgn+i, result.first);	
		}
	}
	return true;	
}

void Occluder::doInsertRegion(Region& rgn)
{
	RegionSet::iterator it;
	Region temp;
	pair<RegionSet::iterator, bool> result;

	result=regionSet.insert(rgn);
	if (!result.second)
	{
		it=result.first;
		if (rgn.rect==it->rect)
		{
			rgn.maxz=min(rgn.maxz, it->maxz);
			regionSet.erase(it);
			regionSet.insert(rgn);
		}
		else
		{	
			fprintf(log, "begin intersect region.......\n");
			fflush(log);
			intersectRegion(&rgn, it);
		}
	}
}

	
void Occluder::insertRegion()
{
	Region* ptr;
	Region rgn;
	RegionSet::iterator it;
	pair<RegionSet::iterator, bool> result;
	for (int i=0; i<TesterCount; i++)
	{
		ptr=regionBuf+i;
		if (ptr->maxz==1.0)
		{
			continue;
		}
		doInsertRegion(*ptr);
	}			
}

/*
bool Occluder::exchangeRegion()
{
	memcpy(regionBuf+myTesterRank, &myRegion, sizeof(Region));
	MPI_Allgather(regionBuf+myTesterRank, sizeof(Region), MPI_UNSIGNED_CHAR, regionBuf,
                sizeof(Region), MPI_UNSIGNED_CHAR, occluderComm);


	for (int i=0; i<TesterCount; i++)
	{
		if (regionBuf[i].maxz==1.0)
		{
			return false;
		}		

		//fprintf(log, "\n**********************\n");
		//fprintf(log, "%d)rect[%d,%d][%d,%d]depth[%f]\n", i, regionBuf[i].rect.topLeft.x,
		//	regionBuf[i].rect.topLeft.y, regionBuf[i].rect.bottomRight.x,
		//	regionBuf[i].rect.bottomRight.y, regionBuf[i].maxz);
	}
	fflush(log);
	return true;
}
	
 
void Occluder::sendEndExchange()
{
	myRegion.maxz=1.0;
	memcpy(regionBuf+myTesterRank, &myRegion, sizeof(Region));
	MPI_Allgather(regionBuf+myTesterRank, sizeof(Region), MPI_UNSIGNED_CHAR, regionBuf,
                sizeof(Region), MPI_UNSIGNED_CHAR, occluderComm);
}
*/

bool Occluder::exchangeRegion()
{
	myRegion.rect.topLeft.x=myRegion.rect.topLeft.y=0;
	myRegion.maxz=1.0;
	myRegion.rect.bottomRight.x=myRegion.rect.bottomRight.y=ScreenSize-1;
	
	glReadPixels(0, 0, ScreenSize, ScreenSize,
		GL_DEPTH_COMPONENT, GL_FLOAT, depthBufferPacket[myTesterRank].depthBuffer);
		
	memcpy(depthBufferPacket+myTesterRank, &myRegion, sizeof(Region));
	MPI_Allgather(depthBufferPacket+myTesterRank, sizeof(DepthBufferPacket), MPI_UNSIGNED_CHAR, depthBufferPacket,
                sizeof(DepthBufferPacket), MPI_UNSIGNED_CHAR, occluderComm);


	for (int i=0; i<TesterCount; i++)
	{
		if (depthBufferPacket[i].region.maxz==-1.0)
		{
			return false;
		}		

		//fprintf(log, "\n**********************\n");
		//fprintf(log, "%d)rect[%d,%d][%d,%d]depth[%f]\n", i, regionBuf[i].rect.topLeft.x,
		//	regionBuf[i].rect.topLeft.y, regionBuf[i].rect.bottomRight.x,
		//	regionBuf[i].rect.bottomRight.y, regionBuf[i].maxz);
	}
	fflush(log);
	return true;
}
	

void Occluder::mergeDepthBuffer()
{
	int index;
	for (int i=0; i<ScreenSize; i++)
	{
		for (int j=0; j<ScreenSize; j++)
		{
			index=i*ScreenSize+j;
			//initialize
			depthBuffer[index]=depthBufferPacket[0].depthBuffer[index];
			//comparison
			for (int k=1; k<TesterCount; k++)
			{
				if (depthBuffer[index]>depthBufferPacket[k].depthBuffer[index])
				{
					depthBuffer[index]=depthBufferPacket[k].depthBuffer[index];
				}
			}
		}
	}
}

void Occluder::checkDepthBuffer()
{
	int index;
	float* ptr = depthBufferPacket[myTesterRank].depthBuffer;
	glReadPixels(0, 0, ScreenSize, ScreenSize,
		GL_DEPTH_COMPONENT, GL_FLOAT, depthBufferPacket[myTesterRank].depthBuffer);

	fprintf(log, "\n****************************\n");
	for (int i=0; i<ScreenSize; i++)
	{
		for (int j=0; j<ScreenSize; j++)
		{
			index=i*ScreenSize+j;
			//fprintf(log, "[(%d,%d)%f]", j, i, depthBuffer[index]);
			if (depthBufferPacket[myTesterRank].depthBuffer[index] != 1.0)
			{
				fprintf(log, "[(%d,%d),%f]", j, i, depthBufferPacket[myTesterRank].depthBuffer[index]);
			}
			if (depthBuffer[index] > ptr[index])
			{
				fprintf(log, "error!!!!!!!!!![%d,%d][%f]\n", j, i, depthBuffer[index]);
			}

		}
		fprintf(log, "\n");
	}
	fprintf(log, "\n");
	fprintf(log, "\n****************************\n");
	fflush(log);
}

void Occluder::printDepthBuffer()
{
	int index;
	fprintf(log, "\n****************************\n");
	for (int i=0; i<ScreenSize; i++)
	{
		for (int j=0; j<ScreenSize; j++)
		{
			index=i*ScreenSize+j;
			fprintf(log, "[(%d,%d)%f]", j, i, depthBuffer[index]);
			//fprintf(log, "[%f]", depthBufferPacket[myTesterRank].depthBuffer[index]);
		}
		fprintf(log, "\n");
	}
	fprintf(log, "\n");
	fprintf(log, "\n****************************\n");
	fflush(log);
}


void Occluder::printDepthBuffer(Rect rect)
{
	int index;
	fprintf(log, "\n****************************\n");
	for (int i=rect.topLeft.y; i<=rect.bottomRight.y; i++)
	{
		for (int j=rect.topLeft.x; j<=rect.bottomRight.x; j++)
		{
			index=i*ScreenSize+j;
			fprintf(log, "[(%d,%d)%f]", j, i, depthBuffer[index]);
			//fprintf(log, "[%f]", depthBufferPacket[myTesterRank].depthBuffer[index]);
		}
		fprintf(log, "\n");
	}
	fprintf(log, "\n");
	fprintf(log, "\n****************************\n");
	fflush(log);
}

bool invalidateRegion(Rect rect)
{
	return rect.topLeft.x >= 0 && rect.topLeft.y < ScreenSize && 
		rect.bottomRight.x >= 0 && rect.bottomRight.y < ScreenSize;
} 

bool Occluder::checkRegion(Rect rect, float depth)
{
	int index;
	if (! invalidateRegion(rect))
	{
		printf("[%d,%d,%d,%d]\n", rect.topLeft.x, rect.topLeft.y, rect.bottomRight.x, rect.bottomRight.y);
		return false;
	}
		
	for (int y=rect.topLeft.y; y<=rect.bottomRight.y; y++)
	{
		for (int x=rect.topLeft.x; x<=rect.bottomRight.x; x++)
		{
			if (depthBuffer[y*ScreenSize+x]>depth)
			{
				return false;//not excluded
			}
		}
	}
	return true;
}
 
void Occluder::sendEndExchange()
{
	myRegion.maxz=-1.0;
	memcpy(depthBufferPacket+myTesterRank, &myRegion, sizeof(Region));
	MPI_Allgather(depthBufferPacket+myTesterRank, sizeof(DepthBufferPacket), MPI_UNSIGNED_CHAR, depthBufferPacket,
                sizeof(DepthBufferPacket), MPI_UNSIGNED_CHAR, occluderComm);
}

