/**********************************************************************
author: qingzhe huang
date: Jul. 31, 2007
*********************************************************************/
#ifndef OCTREE_H
#define OCTREE_H

#include <cstdio>
#include <cstdlib>
#include <vector>
#include "syscall.h"
#include "config.h"

#ifndef LINUX_VERSION
#include <windows.h>
#else
#define LPVOID void*
#define INVALID_HANDLE_VALUE 0xFFFFFFFF
//#define LONGLONG long long
#endif


using namespace std;

const VType MinOctreeNodeWidth=1.0;
const int NodeArrayLengthIncrement=1024;



#define NVector FVector

typedef unsigned char UCVector[4];
typedef unsigned int UIVector[3];


struct PlyVertex
{
	VVector coord;
	UCVector color;
	NVector normal;
	void display()
	{
		printf("coord[%f,%f,%f]\n", coord[0], coord[1], coord[2]);
	}		
};

static bool equalCoord(VVector left, VVector right)
{
	return left[0]==right[0]&&left[1]==right[1]&&left[2]==right[2];
}

struct PlyFace
{
	unsigned char number;
	int* indexList;
	UCVector color;
};


struct VertexInfo
{
	int sonPos;
	int offset;
};

typedef vector<VertexInfo> HashTableNode;
typedef vector<PlyVertex> VertexVector;


struct FaceStruct
{
	UIVector face;
	unsigned int& operator[](int index)
	{
		return face[index];
	}
	void operator=(UIVector triangle)
	{
		memcpy(face, triangle, sizeof(UIVector));
	}
	void display()
	{
		printf("face[%d,%d,%d]\n", face[0], face[1], face[2]);
	}
};

typedef vector<FaceStruct> FaceVector;

void sortTriangleIndex(unsigned int triangle[3]);

struct InfoStruct
{
	int nodeCount;
	int totalVPageCount;
	int totalFPageCount;
	int nodeSize;
	int maxLeafNodeVertexNumber;
	VType maxEdge;
	VType minEdge;
	VType minX, maxX, minZ, maxZ;
	VVector origin;
	int totalNeighbourCount;
	int maxDepth;
	unsigned int totalRepeatFace;
	unsigned int totalRepeatVertex;
	int leafNodeNumber;
	int maxNodeArrayLength;
	unsigned long int totalVertex;
	unsigned long int totalFace;
	InfoStruct();
	
	static char vertexFileName[MAX_PATH];
	static char faceFileName[MAX_PATH];
	static char fakeColorFileName[MAX_PATH];
	static char infoFileName[MAX_PATH];
	static char octreeFileName[MAX_PATH];
	void display();
	void display(FILE* stream);
};

struct Octree
{

	////////////////////////////////////
	//general configuration data

	static InfoStruct infoStruct;

	static VType operations[8][3];

	enum OctreeNodeEnum
	{ 
		TOP_LEFT_FRONT =2, TOP_LEFT_BACK =3, 
		TOP_RIGHT_BACK =7, TOP_RIGHT_FRONT =6,	
		BOTTOM_LEFT_FRONT =0, BOTTOM_LEFT_BACK =1, 
		BOTTOM_RIGHT_BACK =5, BOTTOM_RIGHT_FRONT =4 
	};

	////////////////////////////////////////////
	//utility temporary data
	static Octree* root;
	static HANDLE vertexHandle, faceHandle,	octreeHandle, infoHandle;
	static HANDLE vertexMapHandle, faceMapHandle;
	//static HANDLE pseudoVertexHandle, pseudoFaceHandle;
	/////////////////////////////////////////////////////////////////////
	
	/////////////////////////////////////////////////////////////////////
	//these are user method
	static void uninitialize(bool readOnly);
	static void initialize(bool readOnly, bool useFake=false);
	static void addToOctree(char* inVertexFileName, char* inFaceFileName);

	static void readInfo(bool readOnly);

	static void addToOctree(VertexVector& inVertexVector, FaceVector& inFaceVector);

	///////////////////////////////////////////////////////////////////////////////
	//utility methods and can be used by user
	static int relativePosition(VVector vPoint, VVector vCtr);
	static void setSonNodeCentre(int parentIndex, int sonIndex, int nodeID);
	static int createNewOctreeNode(int myOwnIndex, int nodeID);

	static void initOctreeNode(int nodeIndex);
	static bool isCoordInsideBox(int nodeIndex, VVector coord);
	static void setFilePointerF(int nodeIndex, int offset);
	static void setFilePointerV(int nodeIndex, int offset);
	static void calculateFileOffsetF(int startPageIndex, int offset, unsigned long& low, unsigned long& high);
	static void calculateFileOffsetV(int startPageIndex, int offset, unsigned long& low, unsigned long& high);
	static void calculateEdges(FaceStruct& tri, VertexVector& vertexVector);
	static VType calcEdge(VVector v1, VVector v2);
	////////////////////////////////////////////////////////////////////////////////////////////
	//internal methods and should never be called by user
	static void readDataFromFile(HANDLE vertexHandle, VertexVector& vertexVector);
	static void readDataFromFile(HANDLE faceHandle, FaceVector& vertexVector);
	static void addVertexToOctree(HANDLE outVertexHandle, VertexVector& vertexVector, int nodeIndex);
	static void addFaceToOctree(HANDLE outFaceHandle, FaceVector& faceVector, int nodeIndex);
	static void addFaceToSonFaceVector(HashTableNode* hashTable, FaceStruct& triangle, 
		VertexVector& vertexVector, VertexVector& repeatVertexVector, 
		VertexVector* sonVertexVector[8], VertexVector* sonRepeatVertexVector[8], 
		FaceVector* sonFaceVector[8], FaceVector* sonRepeatFaceVector[8]);

	static void readOriginalFace(int nodeIndex, FaceVector& faceVector, 
			FaceVector& repeatFaceVector, int insertVertexCount, int insertRepeatVertexCount);
	static void readOriginalVertex(int nodeIndex, VertexVector& vertexVector, 
								VertexVector& repeatVertexVector, FaceVector& repeatFaceVector);
	static void expandRootNodeBounding(VVector coord, bool freeExpand);
	static void doAddToOctree(int nodeIndex, VertexVector& vertexVector, 
		VertexVector& repeatVertexVector, FaceVector& faceVector, FaceVector& repeatFaceVector);

	static void calculateNeighbour();
	static void doCalculateNeighbour(int nodeIndex, int neighbourIndex);
	static bool addVertexInfoToHashTable(HashTableNode* hashTable, int vIndex, int sonPos, 
									  int& offset);
	static void doCreateLeafNode(int nodeIndex, VertexVector& vertexVector, 
		VertexVector& repeatVertexVector, FaceVector& faceVector, FaceVector& repeatFaceVector);
	static void distributeVertex(int nodeIndex, HashTableNode* hashTable, VertexVector& vertexVector, 
				VertexVector& repeatVertexVector, VertexVector* sonVertexVector[8]);

	static void doAddToOctreeNodePreparation(HashTableNode*& hashTable, int hashTableSize, VertexVector* sonVertexVector[8], 
		VertexVector* sonRepeatVertexVector[8], FaceVector* sonFaceVector[8], FaceVector* sonRepeatFaceVector[8]);

	static void doAddToOctreeWrapper(int nodeIndex, int vertexCount, int repeatVertexCount, int faceCount, 
		int repeatFaceCount, HANDLE vertexHandle, HANDLE faceHandle);
	//internal methods end
	/////////////////////////////////////////////////////////////////////////////

	static void sortTriangle(FaceStruct& triangle);
	void display();

	////////////////////////////////////////////////////
	//the following are all data member of octree node, some are only meaningful for leaf node
	VVector center;//0
	VType width;//The full width//3
	int vNumber;//4
	int vRepeatNumber;//5
	int fNumber;      //6
	int fRepeatNumber;//7
	int myNodeIndex;  //8
	int parentIndex;  //9
	int sons[8];      //10
	int vStartPageIndex;//vertex//18
	int vPageCount;	         //19
	int fStartPageIndex;//face //20
	int fPageCount;        //21
	unsigned long vFileAddLow;      //22
	unsigned long vFileAddHigh;		//23
	unsigned long fFileAddLow;		//24
	unsigned long fFileAddHigh;		//25
	unsigned int vMapSize;			//26
	unsigned int fMapSize;			//27
	LPVOID vAddress;		//28
	LPVOID fAddress;		//29
	unsigned char sonCount;
	unsigned char level;
	unsigned char neighbourCount;
	int neighbours[MaxNeighbourCount];//maximum 26 neighbours

};

#endif

